/* -------------------------------------------------------------------	*/
/* Copyright (C) 2000 Robert R. Bruner.  All rights reserved.		*/
/* Contact the author at						*/
/*                    rrb@math.wayne.edu				*/
/*                    Dept of Mathematics				*/
/*                    Wayne State University				*/
/*                    Detroit, Michigan  48202				*/
/*                    USA						*/
/*									*/
/* Version 1.6								*/
/* -------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>

#include <errors.h>
#include <structur.h>
#include <storage.h>
#include <steen.h>
#include <io.h>

/* -------------------------------------------------------------------- */
/*  Elements                                                            */
/*  --------                                                            */
/*                                                                      */
/* The linear form of an element (as in a file) is:                     */
/*              #gen gen0 op0 gen1 op1 ... genk opk                     */
/* where k = #gen - 1.  Each op has the form                            */
/*              degree length terms                                     */
/* where degree is an integer, length is unsigned and equals            */
/* dimA(degree), and terms is a sum of operations of the indicated      */
/* degree represented in one of the following forms:                    */
/*                                                                      */
/* binary -                                                             */
/*       b followed by the bytes that form the bitmask.  This is        */
/*       easily read  or written by machine, but is mostly unprintable. */
/* hexadecimal -                                                        */
/*       x followed by the hex representation of the bytes that form    */
/*       the bitmask.                                                   */
/* sequence numbers -                                                   */
/*       s followed by integers between 0 and length-1, separated by    */
/*       commas and terminated with a period.  Typically, but not       */
/*       necessarily, in increasing order.                              */
/* internal -                                                           */
/*       i followed by internal representations of elements enclosed    */
/*       in parentheses, followed by a period, as in                    */
/*                   i(1,0,1)(8).                                       */
/*       meaning Sq(1,0,1) + Sq(8)                                      */
/*                                                                      */
/* ------------------------------------------------------------------   */


/* ------------------------- read_TERM ------------------------------   */

TERM * read_TERM(fp,gen,deg,length)  
FILE *fp;
unsigned gen,length;
int deg;
{
int j,b;
unsigned len;
TERM *t;
char type[2];

    len = 1 + ((length - 1) >> 3);       /* len is in bytes, length in bits  */
    t = get_TERM(length);
    t->gen = gen;
    t->deg = deg;
    if (fscanf(fp,"%1s",type) == EOF)
         terminate("Premature EOF in read_TERM",BAD_READ);
    switch (type[0]) {

    case 'b' : for (j=0; j<len; j++)
                  { int c;
		    if ((c = getc(fp)) == EOF) 
                        terminate("Premature EOF in read_TERM",BAD_READ);
		    t->op[j] = c;
                  }
               break;

    case 'x' : for (j=0; j<len; j++)
                 { int c;
		   if (fscanf(fp,"%2x",&c) == EOF) 
                        terminate("Premature EOF in read_TERM",BAD_READ);
		   t->op[j] = c;
                 }
               break;

    case 's' : for (j=0; j<len; j++) t->op[j] = 0;
	       { char c[2];
		 int res;
               for (c[0]=','; c[0] != '.';) 
                  { if ((res = fscanf(fp,"%d%1s",&b,c)) == EOF) 
                       terminate("Premature EOF in read_TERM",BAD_READ);
                    if (res == 0) 
		      if (fscanf(fp,"%1s",c) == EOF) 
                        terminate("Premature EOF in read_TERM",BAD_READ);
		      else if (c[0] == '.') break;
			else terminate("Malformed op of type s in read_TERM",
				       BAD_READ);
                    t->op[b>>3] |= (0x80) >> (b&7);
                  }
	       }
               break;

    case 'i' : for (j=0; j<len; j++) t->op[j] = 0;
               if (fscanf(fp,"%1s",type) == EOF) 
                   terminate("Premature EOF in read_TERM",BAD_READ);
               while (type[0] == '(' )
                   { Internal_Op O;
		     char c[2];
                     for (j=0,fscanf(fp,"%d%1s",&O[j++],c);
                             c[0] == ',' && j<N_ALG;
                                fscanf(fp,"%d%1s",&O[j++],c)) ;
                      while ( j<N_ALG ) O[j++]=0;
                      b = Seqno(O);
                      t->op[b>>3] |= (0x80) >> (b&7);
                      if (fscanf(fp,"%1s",type) == EOF) 
                          terminate("Premature EOF in read_TERM",BAD_READ);
                   }
               break;


    default :  fprintf(stderr,"\nBad operation type %s. ",type);
               terminate("read_TERM",BAD_READ);

} /* end switch */

return(t);
}

/* ---------------------- read_Element ------------------------------   */

Element * read_Element(fp)  FILE *fp;
{
int  i,deg;
unsigned gen,length,num;
Element *x;

if (fscanf(fp,"%u",&num) == EOF) return(NULL);
          /* EOF here just means there are no more elements to read so   */
	  /*  we return NULL to indicate that.                           */

x = get_Element(num);
for (i=0; i<num; i++) 
 { if (fscanf(fp,"%u %d %u",&gen,&deg,&length) == EOF) 
       terminate("Premature EOF in read_Element",BAD_READ);
          /* EOF here is in mid element so is an error                   */
    x->item[i] = read_TERM(fp,gen,deg,length);
 }

return(x);
}


/* ---------------------- read_exp_Element ------------------------------   */

int read_exp_Element(fp,x,dim)  FILE *fp; TERM *x[]; int dim;
{
   /*      read an element into the expanded version x of an Element        */
   /*      return 1 if successful, EOF if not                               */
   /*      assume x[i] == NULL for all i upon entry, so all we need to do   */
   /*         is replace entries that we read                               */
int  i,deg;
unsigned gen,length,num;
char message[100];

if (fscanf(fp,"%u",&num) == EOF) return(EOF);
          /* EOF here just means there are no more elements to read so   */
	  /*  we return EOF to indicate that.                           */

for (i=0; i<num; i++) 
 { if (fscanf(fp,"%u %d %u",&gen,&deg,&length) == EOF) 
        terminate("Premature EOF in read_exp_Element",BAD_READ);
          /* EOF here is in mid element so is an error                   */
   if (gen >= dim) 
      { sprintf(message,
                "\nread_exp_Element found gen=%d too large for dim=%d",
                gen,dim);
        terminate(message,BAD_READ);
      }
   x[gen] = read_TERM(fp,gen,deg,length);
 } 

return(1);   /* count of elements read, but would 0 be a better value? */
}


/* ---------------------- skip_Element ------------------------------   */

void skip_Element(fp)  FILE *fp;
{
int  i,deg,j;
unsigned gen,len,length,num;
char c[2],type[2];

if (fscanf(fp,"%u",&num) == EOF) return;
          /* EOF here just means there are no more elements to read so   */
	  /*  we simply return                                           */

for (i=0; i<num; i++) 
 {  if (fscanf(fp,"%u %d %u %1s",&gen,&deg,&length,type) == EOF) 
       terminate("Premature EOF in skip_Element",BAD_READ);
          /* EOF here is in mid element so is an error                   */
    len = 1 + ((length - 1) >> 3);   /* len is in bytes, length in bits  */
    switch (type[0]) {

    case 'b' : for (j=0; j<len; j++)
                  if (getc(fp) == EOF) 
                    terminate("Premature EOF in skip_Element",BAD_READ);
                            /* EOF here is in mid element so is an error */
               break;

    case 'x' : for (j=0; j<len; j++)
                  if (fscanf(fp,"%*2x") == EOF) 
                    terminate("Premature EOF in skip_Element",BAD_READ);
                            /* EOF here is in mid element so is an error */
               break;

    case 's' :
               for (c[0]=','; c[0] != '.';) 
                 if (fscanf(fp,"%*d%1s",c) == EOF) 
                     terminate("Premature EOF in skip_Element",BAD_READ);
                            /* EOF here is in mid element so is an error */
               break;

    case 'i' :
               if (fscanf(fp,"%1s",type) == EOF) 
                    terminate("Premature EOF in skip_Element",BAD_READ);
                            /* EOF here is in mid element so is an error */
               while (type[0] == '(' )
                   {
                     for (j=0,fscanf(fp,"%*d%1s",c);
                             c[0] == ',' && j<N_ALG;
                                fscanf(fp,"%*d%1s",c),j++) ;
                      if (fscanf(fp,"%1s",type) == EOF) 
                         terminate("Premature EOF in skip_Element",BAD_READ);
                            /* EOF here is in mid element so is an error */
                   }
               break;


    default :  fprintf(stderr,"\nBad operation type %s. ",type);
               terminate("skip_Element",BAD_READ);

} /* end switch */
} /* end "for i" */

return;
}

/* ------------------------------------------------------------------   */
/* For output, the type parameter can be one of the formats, b,i,s,     */
/*   or x, or it can be 'c', meaning 'condensed', in which case each    */
/*   term will be written in either s or x format, whichever is shorter */

unsigned int trim[8] = {0xff, 0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc, 0xfe};

/* ---------------------- write_TERM -------------------------------   */

write_TERM(fp,x,type) FILE *fp;
                      TERM *x;
                      char type;
{ int j,length,ls,lx,s,n;
  char c;
            /* write_TERM starts writing at the current position, and  */
	    /*  writes a single newline, at the end of the TERM        */

fprintf(fp,"%u %d %u ",x->gen,x->deg,x->len);
length = (x->len - 1) >> 3;
x->op[length++] &= trim[x->len &7];

if (type == 'c')
{  lx = 2*length;
   for (ls=x->op[0]>>7, s=0; (ls <lx) && (s < x->len); s++)
   if (x->op[s>>3] & (0x80 >> (s&7)))
      for (ls++, n=s;  n > 0; ls++, n/=10);
  /* now lx is the length in chars of the hex representation  and  */
  /* ls is the length in chars of the serial number representation */
  /* (omitting the x or s)                                         */

   if ((ls > 0) && (ls < lx)) type = 's'; else type = 'x';
  }


switch (type) {

  case 'b' :
	     fprintf(fp,"b");
	     for  (j=0; j < length; j++) putc(x->op[j],fp);
             fprintf(fp,"\n");
             break;

  case 'i' :
	     {Internal_Op O;
              int seq=0;
	     
	      fprintf(fp,"i");
              for  (j=0,first_op(O,x->deg); j < length; j++) 
               { register unsigned mask = 0x80,
                                     xj = 255 & x->op[j];
				int   k;
                 for (k=0; (xj >0); k++, mask >>= 1)
                  if (xj & mask) { adv_op(O,k+8*j-seq);
                                   fprint_op(fp,O);
                                   seq = k+8*j;
                                   xj ^= mask;
                                 }
               }
             fprintf(fp,".\n");
             }
             break;

  default  :
  case 'x' :
	     fprintf(fp,"x");
	     for  (j=0; j < length; j++) fprintf(fp,"%02x",255 & x->op[j]);
             fprintf(fp,"\n");
             break;

  case 's' :
	     c = 's';
             for  (j=0; j < length; j++) 
              { register unsigned mask = 0x80,
                                     k = 0,
                                    xj = 255 & x->op[j];
                for (; (xj >0); k++, mask >>= 1)
                  if (xj & mask) { fprintf(fp,"%c%d",c,k+8*j);
                                   c = ',';
                                   xj ^= mask;
                                 }
              }
	     if (c == 's') { fprintf(fp,"s ");
			     fprintf(stderr,
			       "\n Warning: printing zero op.\n");
			   }
             fprintf(fp,".\n");
             break;
    } /* end switch */

return(0);
}


/* ---------------------- write_Element ------------------------------   */

write_Element(fp,e,type) FILE *fp;
                         Element *e;
                         char type;
{ int i;

if (e == NULL) { fprintf(fp,"\n0\n"); return(0); }

fprintf(fp,"\n%d\n",e->len);
for (i=0; i < e->len; i++) write_TERM(fp,e->item[i],type);

return(0);
}


/* ---------------------- write_exp_Element ------------------------   */

write_exp_Element(fp,e,type,dim) FILE *fp;
                                 TERM *e[];
                                 char type;
                                 int dim;
{ int i,num;

if (e == NULL) { fprintf(fp,"\n0\n"); return(0); }

/* Determine the number of nonzero entries                             */

for (i=0,num=0; i<dim; i++) if (is_not_zero(e[i])) num++;

fprintf(fp,"\n%d\n",num);
for (i=0; i < dim; i++)
         if (is_not_zero(e[i])) 
		     write_TERM(fp,e[i],type);
return(0);
}
